from typing import List

import attr

from applitools.common.geometry import RectangleSize
from applitools.common.utils import argument_guard
from applitools.common.visual_grid import (
    ChromeEmulationInfo,
    RenderBrowserInfo,
    ScreenOrientation,
)

from .configuration import Configuration
from .misc import BrowserType, StitchMode

__all__ = ("SeleniumConfiguration",)


@attr.s
class SeleniumConfiguration(Configuration):
    DEFAULT_WAIT_BEFORE_SCREENSHOTS_MS = 1000  # ms

    force_full_page_screenshot = attr.ib(default=False)  # type: bool
    wait_before_screenshots = attr.ib(
        default=DEFAULT_WAIT_BEFORE_SCREENSHOTS_MS
    )  # type: int  # ms
    stitch_mode = attr.ib(default=StitchMode.Scroll)  # type: StitchMode
    hide_scrollbars = attr.ib(default=False)  # type: bool
    hide_caret = attr.ib(default=False)  # type: bool

    # Rendering Configuration
    is_throw_exception_on = False
    is_rendering_config = False
    _browsers_info = attr.ib(init=False, factory=list)  # type: List[RenderBrowserInfo]

    def add_browser(self, arg1, arg2=None, arg3=None):
        if isinstance(arg1, RenderBrowserInfo):
            self._browsers_info.append(arg1)
        elif (
            isinstance(arg1, int)
            and isinstance(arg2, int)
            and isinstance(arg3, BrowserType)
        ):
            self._browsers_info.append(
                RenderBrowserInfo(
                    RectangleSize(arg1, arg2), arg3, self.baseline_env_name
                )
            )
        else:
            raise ValueError("Unsupported parameters")
        return self

    def add_device_emulation(self, device_name, orientation=ScreenOrientation.PORTRAIT):
        argument_guard.not_none(device_name)
        emu = ChromeEmulationInfo(device_name, orientation)
        self.add_browser(RenderBrowserInfo(emulation_info=emu))
        return self

    @property
    def browsers_info(self):
        if self._browsers_info:
            return self._browsers_info
        browser_info = RenderBrowserInfo(
            viewport_size=self.viewport_size,
            browser_type=BrowserType.CHROME,
            baseline_env_name=self.baseline_env_name,
        )
        return [browser_info]

    @staticmethod
    def all_fields():
        return list(attr.fields_dict(SeleniumConfiguration).keys())



from __future__ import absolute_import

import hashlib
import itertools
import time
import typing
from datetime import timedelta, tzinfo

from applitools.common import logger

from .compat import parse_qs, urlencode, urlparse, urlsplit, urlunsplit

"""
General purpose utilities.
"""


if typing.TYPE_CHECKING:
    from typing import Callable, Any, List, Text

    T = typing.TypeVar("T")


class _UtcTz(tzinfo):
    """
    A UTC timezone class which is tzinfo compliant.
    """

    _ZERO = timedelta(0)

    def utcoffset(self, dt):
        return _UtcTz._ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return _UtcTz._ZERO


# Constant representing UTC
UTC = _UtcTz()


def use_default_if_none_factory(default_obj, obj):
    def default(attr_name):
        val = getattr(obj, attr_name)
        if val is None:
            return getattr(default_obj, attr_name)
        return val

    return default


def cached_property(f):
    # type: (Callable) -> Any
    """
    Returns a cached property that is calculated by function f
    """

    def get(self):
        try:
            return self._property_cache[f]
        except AttributeError:
            self._property_cache = {}
            x = self._property_cache[f] = f(self)
            return x
        except KeyError:
            x = self._property_cache[f] = f(self)
            return x

    return property(get)


def is_absolute_url(url):
    return bool(urlparse(url).netloc)


def is_url_with_scheme(url):
    return bool(urlparse(url).scheme)


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()

        if "log_time" in kw:
            name = kw.get("log_name", method.__name__.upper())
            kw["log_time"][name] = int((te - ts) * 1000)
        else:
            logger.debug("%r  %2.2f ms" % (method.__name__, (te - ts) * 1000))
        return result

    return timed


def retry(delays=(0, 1, 5), exception=Exception, report=lambda *args: None):
    """
    This is a Python decorator which helps implementing an aspect oriented
    implementation of a retrying of certain steps which might fail sometimes.
    https://code.activestate.com/recipes/580745-retry-decorator-in-python/
    """

    def wrapper(function):
        def wrapped(*args, **kwargs):
            problems = []
            for delay in itertools.chain(delays, [None]):
                try:
                    return function(*args, **kwargs)
                except exception as problem:
                    problems.append(problem)
                    if delay is None:
                        report("retryable failed definitely:", problems)
                        raise
                    else:
                        report(
                            "retryable failed:", problem, "-- delaying for %ds" % delay
                        )
                        time.sleep(delay)

        return wrapped

    return wrapper


def get_sha256_hash(content):
    m = hashlib.sha256()
    m.update(content)
    return "".join(["%02x" % b for b in m.digest()])


def set_query_parameter(url, param_name, param_value):
    """Given a URL, set or replace a query parameter and return the
    modified URL.

    >>> set_query_parameter('http://example.com?foo=bar&biz=baz', 'foo', 'stuff')
    'http://example.com?foo=stuff&biz=baz'

    """
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = parse_qs(query_string)

    query_params[param_name] = [param_value]
    new_query_string = urlencode(query_params, doseq=True)

    return urlunsplit((scheme, netloc, path, new_query_string, fragment))


def proxy_to(proxy_obj_name, fields):
    # type: (Text, List[Text]) -> Callable
    """
    Adds to decorated class __getter__ and __setter__ methods that allow to access
    attributes from proxy_object in the parent class

    :param proxy_obj_name: The name of the proxy object that has decorated class.
    :param fields:
        Fields which should be accessible in parent object from the proxy object.
    """

    def __getattr__(self, name):
        if name in fields:
            proxy_obj = getattr(self, proxy_obj_name)
            return getattr(proxy_obj, name)
        raise AttributeError("{} has not attr {}".format(self.__class__.__name__, name))

    def __setattr__(self, key, value):
        if key in fields:
            proxy_obj = getattr(self, proxy_obj_name)
            setattr(proxy_obj, key, value)
        else:
            super(self.__class__, self).__setattr__(key, value)

    def dec(cls):
        cls.__getattr__ = __getattr__
        cls.__setattr__ = __setattr__
        return cls

    return dec



import os
import uuid
from copy import copy
from datetime import datetime
from typing import Optional, Text, Union

import attr

from applitools.common.geometry import RectangleSize
from applitools.common.match import ImageMatchSettings, MatchLevel
from applitools.common.server import FailureReports, SessionType
from applitools.common.utils import argument_guard, general_utils
from applitools.common.utils.json_utils import JsonInclude

__all__ = ("BatchInfo", "Configuration")

MINIMUM_MATCH_TIMEOUT_MS = 600
DEFAULT_TIMEOUT_MS = 60 * 5 * 1000
DEFAULT_SERVER_URL = "https://eyesapi.applitools.com"


@attr.s
class BatchInfo(object):
    """
    A batch of tests.
    """

    name = attr.ib(
        factory=lambda: os.getenv("APPLITOOLS_BATCH_NAME"),
        metadata={JsonInclude.THIS: True},
    )  # type: Optional[Text]
    started_at = attr.ib(
        factory=lambda: datetime.now(general_utils.UTC),
        metadata={JsonInclude.THIS: True},
    )  # type: Union[datetime, Text]
    sequence_name = attr.ib(
        init=False,
        factory=lambda: os.getenv("APPLITOOLS_BATCH_SEQUENCE"),
        metadata={JsonInclude.NAME: "batchSequenceName"},
    )  # type: Optional[Text]
    id = attr.ib(
        init=False,
        factory=lambda: os.getenv("APPLITOOLS_BATCH_ID", str(uuid.uuid4())),
        metadata={JsonInclude.THIS: True},
    )  # type: Text

    def with_batch_id(self, id):
        argument_guard.not_none(id)
        self.id = id
        return self


@attr.s
class Configuration(object):
    DEFAULT_MATCH_TIMEOUT_MS = 2000

    batch = attr.ib(default=None)  # type: Optional[BatchInfo]
    branch_name = attr.ib(
        factory=lambda: os.getenv("APPLITOOLS_BRANCH", None)
    )  # type: Optional[Text]
    parent_branch_name = attr.ib(
        factory=lambda: os.getenv("APPLITOOLS_PARENT_BRANCH", None)
    )  # type: Optional[Text]
    baseline_branch_name = attr.ib(
        factory=lambda: os.getenv("APPLITOOLS_BASELINE_BRANCH", None)
    )  # type: Optional[Text]
    agent_id = attr.ib(default=None)  # type: Optional[Text]
    baseline_env_name = attr.ib(default=None)  # type: Optional[Text]
    environment_name = attr.ib(default=None)  # type: Optional[Text]
    save_diffs = attr.ib(default=None)  # type: bool
    app_name = attr.ib(default=None)  # type: Optional[Text]
    test_name = attr.ib(default=None)  # type: Optional[Text]
    viewport_size = attr.ib(
        default=None, converter=attr.converters.optional(RectangleSize.from_)
    )  # type: Optional[RectangleSize]
    session_type = attr.ib(default=SessionType.SEQUENTIAL)  # type: SessionType
    ignore_baseline = attr.ib(default=None)  # type: Optional[bool]
    ignore_caret = attr.ib(default=False)
    compare_with_parent_branch = attr.ib(default=None)  # type: Optional[bool]
    host_app = attr.ib(default=None)
    host_os = attr.ib(default=None)
    properties = attr.ib(factory=list)
    hide_scrollbars = attr.ib(default=False)
    match_timeout = attr.ib(default=DEFAULT_MATCH_TIMEOUT_MS)  # ms
    match_level = attr.ib(default=MatchLevel.STRICT, converter=MatchLevel)
    is_disabled = attr.ib(default=False)
    save_new_tests = attr.ib(default=True)
    save_failed_tests = attr.ib(default=False)
    fail_on_new_test = attr.ib(default=False)
    failure_reports = attr.ib(default=FailureReports.ON_CLOSE)
    send_dom = attr.ib(default=True)
    use_dom = attr.ib(default=False)
    enable_patterns = attr.ib(default=False)
    default_match_settings = attr.ib(default=ImageMatchSettings())
    hide_caret = attr.ib(init=False, default=None)
    stitching_overlap = attr.ib(init=False, default=50)

    api_key = attr.ib(factory=lambda: os.getenv("APPLITOOLS_API_KEY", None))
    server_url = attr.ib(default=DEFAULT_SERVER_URL)
    timeout = attr.ib(default=DEFAULT_TIMEOUT_MS)  # ms

    @match_timeout.validator
    def validate1(self, attribute, value):
        if 0 < value < MINIMUM_MATCH_TIMEOUT_MS:
            raise ValueError(
                "Match timeout must be at least {} ms.".format(MINIMUM_MATCH_TIMEOUT_MS)
            )

    @viewport_size.validator
    def validate2(self, attribute, value):
        if value is None:
            return None
        if not isinstance(value, RectangleSize) or not (
            isinstance(value, dict)
            and "width" in value.keys()
            and "height" in value.keys()
        ):
            raise ValueError("Wrong viewport type settled")

    @property
    def is_dom_send(self):
        return self.send_dom

    def clone(self):
        return copy(self)

    @property
    def short_description(self):
        return "{} of {}".format(self.test_name, self.app_name)

    @staticmethod
    def all_fields():
        return list(attr.fields_dict(Configuration).keys())



from __future__ import absolute_import

import typing
from enum import Enum

import attr

from .geometry import RectangleSize
from .match import ImageMatchSettings

if typing.TYPE_CHECKING:
    from typing import Text, Any, Dict, Optional, List

    # TODO: Implement objects
    SessionUrls = Dict[Any, Any]
    StepInfo = Dict[Any, Any]

__all__ = ("TestResults", "TestResultSummary")


class TestResultsStatus(Enum):
    """
    Status values for tests results.
    """

    Passed = "Passed"
    Unresolved = "Unresolved"
    Failed = "Failed"


@attr.s
class TestResults(object):
    """
    Eyes test results.
    """

    steps = attr.ib(default=0)  # type: int
    matches = attr.ib(default=0)  # type: int
    mismatches = attr.ib(default=0)  # type: int
    missing = attr.ib(default=0)  # type: int
    exact_matches = attr.ib(default=0, repr=False)  # type: int
    strict_matches = attr.ib(default=0, repr=False)  # type: int
    content_matches = attr.ib(default=0, repr=False)  # type: int
    layout_matches = attr.ib(default=0, repr=False)  # type: int
    none_matches = attr.ib(default=0, repr=False)  # type: int
    is_new = attr.ib(default=None, repr=False)  # type: Optional[bool]

    url = attr.ib(init=False, default=None)  # type: Optional[Text]
    status = attr.ib(default=None, repr=False)  # type: Optional[Text]

    name = attr.ib(default=None, repr=False)  # type: Text
    secret_token = attr.ib(default=None, repr=False)  # type: Text
    id = attr.ib(default=None, repr=False)  # type: Text
    app_name = attr.ib(default=None, repr=False)  # type: Text
    batch_name = attr.ib(default=None, repr=False)  # type: Text
    batch_id = attr.ib(default=None, repr=False)  # type: Text
    branch_name = attr.ib(default=None, repr=False)  # type: Text
    host_os = attr.ib(default=None, repr=False)  # type: Text
    host_app = attr.ib(default=None, repr=False)  # type: Text
    host_display_size = attr.ib(default=None, repr=False)  # type: RectangleSize
    # started_at = attr.ib(
    #     default=None,
    #     repr=False,
    #     converter=lambda d: datetime.strptime(d, "%Y-%m-%dT%H:%M:%S.%fZ"),
    # )  # type: datetime
    started_at = attr.ib(default=None, repr=False)  # type: Text
    duration = attr.ib(default=None, repr=False)  # type: int
    is_different = attr.ib(default=None, repr=False)  # type: bool
    is_aborted = attr.ib(default=None, repr=False)  # type: bool
    app_urls = attr.ib(default=None, repr=False)  # type: SessionUrls
    api_urls = attr.ib(default=None, repr=False)  # type: SessionUrls
    steps_info = attr.ib(default=None, repr=False)  # type: StepInfo
    baseline_id = attr.ib(default=None, repr=False)  # type: Text
    default_match_settings = attr.ib(default=None, repr=False, type=ImageMatchSettings)

    @property
    def is_passed(self):
        # type: () -> bool
        return (
            self.status is not None
        ) and self.status.lower() == TestResultsStatus.Passed.name.lower()

    @property
    def is_unresolved(self):
        # type: () -> bool
        return (
            self.status is not None
        ) and self.status.lower() == TestResultsStatus.Unresolved.name.lower()

    @property
    def is_failed(self):
        # type: () -> bool
        return (
            self.status is not None
        ) and self.status.lower() == TestResultsStatus.Failed.name.lower()

    def __str__(self):
        origin_str = super(TestResults, self).__str__()
        preamble = "New test" if self.is_new else "Existing test"
        return "{} [{}]".format(preamble, origin_str)


@attr.s
class TestResultSummary(object):
    all_results = attr.ib()  # type: List[TestResults]
    exceptions = attr.ib(default=0)

    passed = attr.ib(init=False, default=0)
    unresolved = attr.ib(init=False, default=0)
    failed = attr.ib(init=False, default=0)
    mismatches = attr.ib(init=False, default=0)
    missing = attr.ib(init=False, default=0)
    matches = attr.ib(init=False, default=0)

    def __attrs_post_init__(self):
        for result in self.all_results:
            if result is None:
                continue
            if result.is_failed:
                self.failed += 1
            elif result.is_passed:
                self.passed += 1
            elif result.is_unresolved:
                self.unresolved += 1
            self.matches += result.matches
            self.missing += result.missing
            self.mismatches += result.mismatches



import enum
import json
import re
from datetime import datetime

import attr

from .compat import iteritems


def to_json(val):
    return json.dumps(val, default=to_serializable, sort_keys=True)


def _fields_from_attr(cls):
    return [field.name for field in attr.fields(cls)]


def _klasses_from_attr(cls):
    klasses = {}
    klasses[cls] = _fields_from_attr(cls)
    for field in attr.fields(cls):
        if field.type and attr.has(field.type):
            klasses[field.type] = _fields_from_attr(field.type)
    return klasses


def attr_from_json(content, cls):
    klasses = _klasses_from_attr(cls)

    def make_snake(dct):
        params = {}
        for k, v in iteritems(dct):
            k = camelcase_to_underscore(k)
            # if k in fields:
            params[k] = v
        return params

    def obj_came(obj):
        params = make_snake(dict(obj))
        for kls, fields in iteritems(klasses):
            if len(klasses) == 1:
                cleaned_params = {
                    key: val for key, val in iteritems(params) if key in fields
                }
                return kls(**cleaned_params)

            coincidence = 0
            for key in params.keys():
                if key in fields:
                    coincidence += 1
            if coincidence >= 2:
                cleaned_params = {
                    key: val for key, val in iteritems(params) if key in fields
                }
                return kls(**cleaned_params)

    instance = json.loads(content, object_hook=obj_came)
    return instance


def attr_from_response(response, cls):
    return attr_from_json(response.text, cls)


# Uses for replacing of regular attr.name to specified in metadata
REPLACE_TO_DICT = dict()


class _CamelCasedDict(dict):
    def __setitem__(self, key, value):
        if key in REPLACE_TO_DICT:
            # use key specified in metadata
            old_key = key
            key = REPLACE_TO_DICT[old_key]
            del REPLACE_TO_DICT[old_key]
        else:
            # convert key into camel case format
            key = underscore_to_camelcase(key)
        # process Enum's
        if hasattr(value, "value"):
            value = value.value
        super(_CamelCasedDict, self).__setitem__(key, value)


def _filter(attr_, value):
    if attr_.name.startswith("_"):
        return False
    if attr_.metadata.get(JsonInclude.NON_NONE):
        if value is None:
            return False
        return True
    if attr_.metadata.get(JsonInclude.THIS):
        return True
    if attr_.metadata.get(JsonInclude.NAME):
        # set key from metadata which would be used by default
        REPLACE_TO_DICT[attr_.name] = attr_.metadata[JsonInclude.NAME]
        return True
    return False


def to_serializable(val):
    if isinstance(val, datetime):
        return val.strftime("%Y-%m-%dT%H:%M:%SZ")
    elif isinstance(val, enum.Enum):
        return val.value
    elif attr.has(val.__class__):
        name = getattr(val, "JSON_NAME", None)
        obj = attr.asdict(val, filter=_filter, dict_factory=_CamelCasedDict)
        if name:
            return {name: obj}
        return obj
    elif isinstance(val, bytes):
        return val
    elif isinstance(val, Exception):
        return {"error": val.__class__.__name__, "args": val.args}
    return str(val)


def camelcase_to_underscore(text):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def underscore_to_camelcase(text):
    return re.sub(r"(?!^)_([a-zA-Z])", lambda m: m.group(1).upper(), text)


class JsonInclude(object):
    NON_NONE = "non_none"
    THIS = "include"
    NAME = "name"



from .configuration import BatchInfo, Configuration
from .misc import BrowserType, StitchMode
from .selenium_configuration import SeleniumConfiguration

__all__ = (
    "Configuration",
    "BatchInfo",
    "StitchMode",
    "BrowserType",
    "SeleniumConfiguration",
)



from .app_output import AppOutput
from .capture import EyesScreenshot
from .config import (
    BatchInfo,
    BrowserType,
    Configuration,
    SeleniumConfiguration,
    StitchMode,
)
from .errors import (
    CoordinatesTypeConversionError,
    DiffsFoundError,
    EyesError,
    EyesIllegalArgument,
    NewTestError,
    OutOfBoundsError,
    TestFailedError,
)
from .geometry import EMPTY_REGION, CoordinatesType, Point, RectangleSize, Region
from .logger import FileLogger, NullLogger, StdoutLogger
from .match import (
    ExactMatchSettings,
    FloatingBounds,
    FloatingMatchSettings,
    ImageMatchSettings,
    MatchLevel,
    MatchResult,
)
from .match_window_data import MatchWindowData, Options
from .metadata import AppEnvironment, RunningSession, SessionStartInfo
from .server import FailureReports, SessionType
from .test_results import TestResults, TestResultSummary
from .visual_grid import (
    ChromeEmulationInfo,
    DeviceName,
    EmulationDevice,
    RenderBrowserInfo,
    RenderInfo,
    RenderingInfo,
    RenderRequest,
    RenderStatus,
    RenderStatusResults,
    RGridDom,
    RunningRender,
    ScreenOrientation,
    VGResource,
    VisualGridSelector,
)

__all__ = (
    logger.__all__  # noqa
    + config.__all__  # noqa
    + errors.__all__  # noqa
    + geometry.__all__  # noqa
    + match.__all__  # noqa
    + metadata.__all__  # noqa
    + app_output.__all__  # noqa
    + capture.__all__  # noqa
    + match_window_data.__all__  # noqa
    + test_results.__all__  # noqa
    + visual_grid.__all__  # noqa
    + server.__all__  # noqa
    + ("logger", "StitchMode")  # noqa
)
